(* Not entirely implemented *)

(*################################*)
 IMPLEMENTATION MODULE ArgsSupport;         (* v0.2-051704 for StonyBrook Win32 *)
(*################################*)
                                       (* by Thomas Breeden, tmb@virginia.edu *)
(*
Copyright (C) 1999-2004 Thomas Breeden.

Permission to use, copy, modify and distribute this software and its
documentation for any purpose and without fee is hereby granted, provided that
the above copyright notice appear in all copies and that both the copyright
notice and this permission notice appear in supporting documentation.
Thomas Breeden makes no representations about the suitability of this software
for any purpose. It is provided "as is" without express or implied warranty.
*)

(*
     Support routines for relatively painless specification, reading, and initial
     parsing of command line arguments.

     Specification templates are a la AmigaOS.

     Note that each argument has an associated keyword, which is usually optional
     on the command line (unless it is a switch argument or "/K" is in the template).
     Parameters are filled positionally if the keyword is not specified on the
     command line.

     An example template:

        PgmTemplate = "-f=-fail/K,-p=-pass/K,-script/A,-driver,-overwrite/S,-h=-help/S";

     Specifies that

        The fail and pass parameters are string parameters (the default) and are optional,
        but if used must include their keywords.

        The script parameter is a string parameter and is required (/A = Always).
        If it's keyword is not supplied then the first non-keyworded argument
        is taken as the script name.

        The driver parameter is similar, but not required.

        The overwrite parameter is a switch parameter (/S = Switch).

     Would accept command lines like

        MyProgram test.ts
        MyProgram -fail full -overwrite test.ts test2.mod
        MyProgram -p titles -script test.ts

     Note: The '-' character is not special here. The leading dash in the keywords
     above is a stylistic convention.

     See the .DEF file for more on the specification syntax.
*)

FROM SYSTEM     IMPORT ADR;
FROM Debugging0 IMPORT Assert;

FROM ConvTypes   IMPORT ConvResults;
FROM IOChan      IMPORT ChanId;
FROM IOConsts    IMPORT ReadResults;
FROM IOResult    IMPORT ReadResult;
FROM ProgramArgs IMPORT ArgChan, IsArgPresent, NextArg;
FROM TextIO      IMPORT ReadString, SkipLine;
FROM WholeStr    IMPORT StrToInt;

FROM DynStr0     IMPORT DynStr;
FROM DynStr2A    IMPORT dSkipDelims, (*dStrCollapse,*) dNextToken, WhiteSpace;
IMPORT DynStr0, DynStr1, DynStr1b, DynStr1C, DynStr3, Str0, Str1;

CONST  LF    = 12C;
       QUOTE = '"';
       MaxNumArgs = 20;

TYPE   Switches = (SwitchSw, KeywordSw, NumberSw, ToggleSw, RequiredSw,
                   RestOfLineSw, MultiStringSw);
       SwitchSet = SET OF Switches;
CONST  SwitchNames = "SKNTAFM";

TYPE   ParamsInfoRec = RECORD
                         Key0,
                         Key1        :DynStr;
                         KeyWrdGiven :BOOLEAN;
                         MultiStrCnt :INTEGER;
                         switches    :SwitchSet;
                       END;
(* VERSION HISTORY

v0.0-022299 for StonyBrook Win32
            -first

v0.1-061999 -Added support for /M switch.

    -110399 -A SwitchSw is by definition a "/K" switch, and is now treated as such even if the /K is not
             in the template.

v0.2-121800 -Added quote processing for string parameters.

    -011504 -Fix: We want to leave all quotes alone for /M or /F parameters.
             otherwise, we lose all info about the multiple tokens. So, made sure that
             they are not removed.

             /M still need some work w/r/t quoting if /A arguments need to be "stolen"

    -040504 -Changed name of Proc GetCLI() to ProcessCLI() to agree with
                  Amiga version.

    -051704 -Replaced StrValue.StrInt() with WholeStr.StrToInt()
*)

(*=================================================================*)
 PROCEDURE InitParamEntry(typ:ParamsTypes; VAR pe:ParamsArrayEntry);
(*=================================================================*)
       (* assumes .vs field not created, sets .str to NIL *)
BEGIN WITH pe DO

Val.ptype := typ;
CASE typ OF
   IntParam:     Val.vi := 0; |
   SwitchParam:  Val.vb := FALSE; |
   StrParam:     Val.vs.str := NIL;
END;

Given := FALSE;

END END InitParamEntry;

(*===============================================*)
 PROCEDURE ClrParamEntry(VAR pe:ParamsArrayEntry);
(*===============================================*)
       (* assumes .vs field created or NIL *)
BEGIN WITH pe DO

CASE Val.ptype OF
   IntParam:     Val.vi := 0; |
   SwitchParam:  Val.vb := FALSE; |
   StrParam:     IF Val.vs.str # NIL THEN
                    DynStr1.dStrClr(Val.vs);
                 END;
END;

Given := FALSE;

END END ClrParamEntry;

(*===================================================*)
 PROCEDURE DisposeParamEntry(VAR pe:ParamsArrayEntry);
(*===================================================*)
       (* assumes .vs field created or NIL *)
BEGIN WITH pe DO

CASE Val.ptype OF
   IntParam:     Val.vi := 0; |
   SwitchParam:  Val.vb := FALSE; |
   StrParam:     IF Val.vs.str # NIL THEN
                    DynStr0.dStrDispose(Val.vs);
                 END;
END;

Given := FALSE;

END END DisposeParamEntry;

(*----------------------------------------------------------------------------------------*)
 PROCEDURE MatchKey(token:DynStr; VAR ParamsInfo:ARRAY OF ParamsInfoRec; NumParams:INTEGER;
                    VAR parm:INTEGER):BOOLEAN;
(*----------------------------------------------------------------------------------------*)

VAR  i :INTEGER;

BEGIN

FOR i := 0 TO NumParams-1 DO
   IF (DynStr1.dStrCmp(token, ParamsInfo[i].Key0) = 0)
                  OR
          ((ParamsInfo[i].Key1.str # NIL) AND (DynStr1.dStrCmp(token, ParamsInfo[i].Key1) = 0)) THEN
      ParamsInfo[i].KeyWrdGiven := TRUE;
      parm := i;
      RETURN TRUE;
   END;
END;

RETURN FALSE;

END MatchKey;

(*-----------------------------------------*)
 PROCEDURE FindLastSpace(s:DynStr):INTEGER;
(*-----------------------------------------*)

VAR ret,
    pos,
    Start :INTEGER;

BEGIN

ret := -1;
pos := DynStr1.dStr0Pos(s, " ", 0);

WHILE pos >= 0 DO
   Start := pos+1;
   ret := pos;
   pos := DynStr1.dStr0Pos(s, " ", Start);
END;

RETURN ret;

END FindLastSpace;

(*------------------------------------------------------------------------------------*)
 PROCEDURE AssignStr(token:DynStr; parm:INTEGER; VAR Params:ARRAY OF ParamsArrayEntry;
                     VAR Info:ARRAY OF ParamsInfoRec);
(*------------------------------------------------------------------------------------*)

BEGIN

IF Params[parm].Val.vs.str = NIL THEN

   DynStr0.dStrAsgC(token, Params[parm].Val.vs);

ELSE

   IF (MultiStringSw IN Info[parm].switches) THEN

      IF Info[parm].MultiStrCnt = 0 THEN
         DynStr3.dStrAsgE(token, Params[parm].Val.vs);
      ELSE
         DynStr3.dChrCat(" ", Params[parm].Val.vs);
         DynStr3.dStrCat(token, Params[parm].Val.vs);
      END;

   ELSE

      DynStr3.dStrAsgE(token, Params[parm].Val.vs);

   END;

END;

IF (MultiStringSw IN Info[parm].switches) THEN
   INC(Info[parm].MultiStrCnt);
END;

END AssignStr;

(*==================================================================================================*)
 PROCEDURE GetParams(VAR args:DynStr; template:DynStr; VAR Params:ARRAY OF ParamsArrayEntry):BOOLEAN;
(*==================================================================================================*)

VAR   parm,
      NumParams       :INTEGER;
      MultiStrArg     :INTEGER;
      inx,
      pos             :INTEGER;
      delimsTempl,
      delimsCLI       :Str0.String10;
      ParamsInfoArray :ARRAY [0..MaxNumArgs-1] OF ParamsInfoRec;
      token           :DynStr;
      cnvRes          :ConvResults;
      typSet,
      pFound,
      more,
      quoted,
      bRes            :BOOLEAN;

               (*----------------------*)
                PROCEDURE GetCLIToken();
               (*----------------------*)
               BEGIN

               dNextToken(args, delimsCLI, token, inx);
               more := token.len > 0;
               quoted := more AND (token.str^[0] = QUOTE);
               IF quoted THEN
                  DynStr1.dStrSeg(token, 1, token.len-1, token);
                  more := token.len > 0;
                  IF more AND (token.str^[token.len-1] = QUOTE) THEN
                     DynStr1.dStrSeg(token, 0, token.len-2, token);           (* ok for null string, but wrong for ""ABC *)
                     quoted := FALSE;
                  END;
               END;

               END GetCLIToken;

BEGIN

bRes := TRUE;

FOR parm := 0 TO VAL(INTEGER, HIGH(Params)) DO
   Params[parm].Given := FALSE;
END;

Str0.StrAsg(WhiteSpace, delimsTempl);
Str1.StrCat("=/,", delimsTempl);

inx := 0;
dSkipDelims(template, delimsTempl, inx);
(*dStrCollapse(template, WhiteSpace);*) (* NO DO NOT CHANGE Template! *)

IF inx >= VAL(INTEGER, args.len) THEN
   RETURN (DynStr1.dStr0Pos(template, "/A", 0) < 0)
END;

FOR parm := 0 TO MaxNumArgs-1 DO
   ParamsInfoArray[parm].Key0.str := NIL;
   ParamsInfoArray[parm].Key1.str := NIL;
   ParamsInfoArray[parm].KeyWrdGiven := FALSE;
   ParamsInfoArray[parm].MultiStrCnt := 0;
   ParamsInfoArray[parm].switches := SwitchSet{};
END;

DynStr0.dStrCreate(token, 80);

(* Interpret the template string *)
MultiStrArg := -1;
more := TRUE;
parm := 0;
WHILE more AND bRes DO

   IF parm >= MaxNumArgs THEN
      bRes := FALSE;
   END;

   (* arg names *)
   dNextToken(template, delimsTempl, token, inx);
   DynStr0.dStrAsgC(token, ParamsInfoArray[parm].Key0);
   IF template.str^[inx] = '=' THEN
      dNextToken(template, delimsTempl, token, inx);
      IF token.len = 0 THEN
         bRes := FALSE;
      ELSE
         DynStr0.dStrAsgC(token, ParamsInfoArray[parm].Key1);
      END;
   END;

   typSet := FALSE;
   WHILE template.str^[inx] = '/' DO
      dNextToken(template, delimsTempl, token, inx);
      IF token.len = 0 THEN
         bRes := FALSE;
      END;
      pos := Str1.ChrPos(SwitchNames, token.str^[0], 0);
      IF pos < 0 THEN
         bRes := FALSE;
      ELSE
         INCL(ParamsInfoArray[parm].switches, VAL(Switches, pos));
         CASE VAL(Switches, pos) OF
            SwitchSw:
               IF Params[parm].Val.ptype # SwitchParam THEN
                  bRes := FALSE;
               END;
               IF NOT typSet THEN
                  typSet := TRUE;
                  INCL(ParamsInfoArray[parm].switches, KeywordSw);
               ELSE
                  bRes := FALSE;
               END;
               |
            NumberSw:
               IF Params[parm].Val.ptype # IntParam THEN
                  bRes := FALSE;
               END;
               IF NOT typSet THEN
                  typSet := TRUE;
               ELSE
                  bRes := FALSE;
               END;
               |
            MultiStringSw:
               IF MultiStrArg < 0 THEN
                  MultiStrArg := parm;
               ELSE
                  bRes := FALSE;
               END;
               |
            (*RequiredSw:
               |*)
            ELSE
         END;
      END;
   END;
   IF NOT typSet AND (Params[parm].Val.ptype # StrParam) THEN
      bRes := FALSE;
   END;
   IF inx >= VAL(INTEGER, template.len) THEN
      more := FALSE;
   ELSIF template.str^[inx] # ',' THEN
      bRes := FALSE;
   END;

   INC(parm);
END;
NumParams := parm;

(* process the command line *)
Str0.StrAsg(WhiteSpace, delimsCLI);
Str1.StrCat("=", delimsCLI);

inx := 0;
(* >>> NOT FOR ISO >>>> dNextToken(args, WhiteSpace, token, inx);   (* for Win32, first token is the program name *)*)
(*dSkipDelims(args, delimsCLI, inx);*)

GetCLIToken();

WHILE more AND bRes DO

   pFound := MatchKey(token, ParamsInfoArray, NumParams, parm);

   IF NOT pFound THEN

      IF MultiStrArg >= 0 THEN
         parm := MultiStrArg;
         pFound := TRUE;
         Assert(NOT (NumberSw IN ParamsInfoArray[parm].switches), "/M/N processing NYI");
      ELSE
         (* pick next ungiven and not key necessary parameter *)
         parm := 0;
         WHILE NOT pFound AND (parm < NumParams) DO
            IF (NOT Params[parm].Given) AND NOT (KeywordSw IN ParamsInfoArray[parm].switches) THEN
               pFound := TRUE;
            ELSE
               INC(parm);
            END;
         END;
      END;

   ELSE

      IF Params[parm].Val.ptype # SwitchParam THEN
         dSkipDelims(args, delimsCLI, inx);
      END;

   END;

   IF NOT pFound THEN

      bRes := FALSE;

   ELSE

      Params[parm].Given := TRUE;
      CASE Params[parm].Val.ptype OF

         IntParam:
            IF ParamsInfoArray[parm].KeyWrdGiven THEN
               GetCLIToken();
            END;
            IF quoted OR NOT more THEN
               bRes := FALSE;
            ELSE
               StrToInt(token.str^, Params[parm].Val.vi, cnvRes);
               IF cnvRes # strAllRight THEN
                  bRes := FALSE;
               END;
            END; |

         SwitchParam:

            Params[parm].Val.vb := TRUE; |

         StrParam:

            IF (RestOfLineSw IN ParamsInfoArray[parm].switches) OR (MultiStringSw IN ParamsInfoArray[parm].switches) THEN

               IF ParamsInfoArray[parm].KeyWrdGiven THEN
                  dSkipDelims(args, delimsCLI, inx);
                  DynStr3.dStrSeg(args, inx, args.len-1, token);      (* rest of line after keyword *)
               ELSE
                  DynStr3.dStrSeg(args, inx-VAL(INTEGER, token.len), args.len-1, token);   (* rest of line, no keyword *)
                  IF quoted THEN
                     DynStr3.dStr0Pre('"', token);  (* put the leading quote back *)
                     quoted := FALSE;
                  END;
               END;
               more := FALSE;

            ELSE

               IF ParamsInfoArray[parm].KeyWrdGiven THEN
                  GetCLIToken();
                  IF NOT more THEN
                     bRes := FALSE;
                  END;
               END;

            END;

            IF bRes THEN
               AssignStr(token, parm, Params, ParamsInfoArray);
               IF quoted THEN
                  dNextToken(args, QUOTE, token, inx);
                  more := token.len > 0;
                  bRes := (inx < VAL(INTEGER, args.len)) AND (args.str^[inx] = QUOTE);
                  IF bRes THEN
                     INC(inx);
                     DynStr3.dStrCat(token, Params[parm].Val.vs);
                  END;
               END;
            END; |

      END; (*CASE*)

   END;

   IF more THEN
      GetCLIToken();
   END;
END;

(* steal from MultiString if necessary *)
IF MultiStrArg >= 0 THEN
   FOR parm := 0 TO NumParams-1 DO
      IF (parm # MultiStrArg) AND (ParamsInfoArray[parm].MultiStrCnt > 0)
                AND (RequiredSw IN ParamsInfoArray[parm].switches) AND (NOT Params[parm].Given) THEN
         DEC(ParamsInfoArray[parm].MultiStrCnt);
         Assert(NOT (NumberSw IN ParamsInfoArray[parm].switches), "/M/N processing NYI");
         pos := FindLastSpace(Params[MultiStrArg].Val.vs);
         IF pos >= 0 THEN
            AssignStr(Params[MultiStrArg].Val.vs, parm, Params, ParamsInfoArray);
         ELSE
            DynStr3.dStrSeg(Params[MultiStrArg].Val.vs, pos+1, Params[MultiStrArg].Val.vs.len-1, token);
            DynStr1.dStrSeg(Params[MultiStrArg].Val.vs, 0, pos-1, Params[MultiStrArg].Val.vs);
            AssignStr(token, parm, Params, ParamsInfoArray);
         END;
      END;
   END;
   IF ParamsInfoArray[MultiStrArg].MultiStrCnt = 0 THEN
      Params[MultiStrArg].Given := FALSE;
   END;
END;

FOR parm := 0 TO NumParams-1 DO
   IF (RequiredSw IN ParamsInfoArray[parm].switches) AND (NOT Params[parm].Given) THEN
      bRes := FALSE;
   END;
END;

DynStr0.dStrDispose(token);

RETURN bRes;

END GetParams;

(*=====================================================================================*)
 PROCEDURE ProcessCLI(template:ARRAY OF CHAR; VAR Params:ARRAY OF ParamsArrayEntry):BOOLEAN;
(*=====================================================================================*)

VAR CommandLine,
    Template      :DynStr;
    stemp         :Str0.String80;
    rRes          :ReadResults;
    bRes          :BOOLEAN;

BEGIN

DynStr0.dStrCreate(CommandLine, 80);
DynStr0.dStrInC(template, Template);

WHILE IsArgPresent() DO
   REPEAT
      ReadString(ArgChan(), stemp);
      rRes := ReadResult(ArgChan());
      DynStr3.dStr0Cat(stemp, CommandLine);
   UNTIL rRes # allRight;
   NextArg();
END;

bRes := GetParams(CommandLine, Template, Params);

DynStr0.dStrDispose(Template);
DynStr0.dStrDispose(CommandLine);

RETURN bRes;

END ProcessCLI;

END ArgsSupport.
