
(*############################*)
 IMPLEMENTATION MODULE DynStr3;         (* v1.5-080501 for Benchmark AmigaDOS *)
(*############################*)
                                       (* by Thomas Breeden, tmb@virginia.edu *)
(*
        Dynamically allocated string processing: procedures supporting
                                                 dynamic reallocation.
*)

(*
Copyright (C) 1989-2001 Thomas Breeden.

Permission to use, copy, modify and distribute this software and its
documentation for any purpose and without fee is hereby granted, provided that
the above copyright notice appear in all copies and that both the copyright
notice and this permission notice appear in supporting documentation.
Thomas Breeden makes no representations about the suitability of this software
for any purpose. It is provided "as is" without express or implied warranty.
*)

(*?D-*) (* dynamic array copying *)
(*?R-*)  (* run time checks *)

(*FROM Debugging IMPORT Debug;*)
FROM DynStr0   IMPORT DynStr, dStrDispose, dStrCreate;
IMPORT DynStr0, DynStr1, DynStr1C;

CONST ExpectedLo       = 40;
      ExpectedHi       = 120;
      ExpectedRange = ExpectedHi - ExpectedLo + 1;
      (*ExpectedSize  = ;*)

(*
NOTE


VERSION HISTORY

v1.0-121189     -first

v1.1-042290     -Added str0 concatenate proc.

v1.2-042690     -Fixed problem with input = output for dStrAsgE, dStrSeg. Now
                 checking and preserving value if so.

v1.3-062093     -Fixed SizeCheck use of Hysteresis when the string fits into
                 the size! Before, it was ALWAYS creating a new dynstr. Now
                 it works as intended: only creating a new dynstr if the
                 size exceeds needed by Hysteresis or more.

v1.4-050497     -Added dChrPre() proc.
                -Fixed INT->CARD bug in dStrSeg().
                -Removed "hysteresis". Instead using an "expected size" value
                 and an "expected range" with these rules:
                     > If need to reallocate a bigger siz
                              If needed size is within expected range then allocate the top of expected range.
                              Else allocate the needed size + 1/2 range.
                     > If new size is smaller
                              If was in expected range
                                 If new size is in expected range, no deallocation.
                                 Else deallocate to bottom of expected range.
                              Elsif was above expected range
                                 If new is still above expected range AND decrement is more than expected range then deallocate to newsize + 1/2 expected range
                                 If new is in range, allocate to top of expected range
                              Else (was below expected range)
                                 <same as first case above>

    -062297     -Fixed bug design: in many of these problems occurred if the outstring is also
                 the instring. SizeCheck() can deallocate the outstring to change it's size, but the
                 instring was used afterwards without taking account that it had been deallocated.
                 Fixed by adding a param to SizeCheck to hold the dynstr to be deallocated,
                 and delaying the actual deallocation until later.

v1.5-050399     -Exporting SizeCheck().

    -080501     -Changed all StrLen() to LENGTH()
                -Documented that dStrCat() and dStrAsgE() will work correctly
                 if there are embedded 0C chars in the input DynStr.

*)

(*---------------------------------------------------------------------------------------------------*)
 PROCEDURE SizeCheckNoDealloc(asglen:CARDINAL; VAR s2:DynStr; preserve:BOOLEAN; VAR deallocMe:DynStr);
(*---------------------------------------------------------------------------------------------------*)

VAR  dstrnew     :DynStr;
     NewInRange,
     OldInRange  :BOOLEAN;
     newsiz      :CARDINAL;

BEGIN WITH s2 DO

deallocMe.str := NIL;

newsiz := siz;
NewInRange := (asglen >= ExpectedLo) AND (asglen <= ExpectedHi);

IF siz < asglen THEN

   (* need to enlarge *)

   IF NewInRange THEN
      newsiz := ExpectedHi;
   ELSE
      newsiz := asglen + (ExpectedRange DIV 2);
   END;

ELSIF (siz - asglen) > ExpectedRange THEN

   (* possible can shrink *)

   OldInRange := (siz >= ExpectedLo) AND (siz <= ExpectedHi);

   IF OldInRange THEN
      IF NOT NewInRange THEN
         newsiz := asglen + (ExpectedRange DIV 2);  (* went below range *)
      END;
   ELSIF (siz > ExpectedHi) AND NewInRange THEN        (* fell into range from above *)
      newsiz := ExpectedHi;
   ELSE                                             (* decreased by a big amount *)
      newsiz := asglen + (ExpectedRange DIV 2);
   END;

END;

IF newsiz # siz THEN

   (*>>>>>>>>>>>
   Debug("asglen", asglen);
   Debug("oldsiz", siz);
   Debug("newsiz", newsiz);
   <<<<<<<*)

   dStrCreate(dstrnew, newsiz);
   IF preserve THEN
      DynStr0.dStrAsgE(s2, dstrnew);
   END;
   (*dStrDispose(s2);*)
   deallocMe := s2;
   s2 := dstrnew;

END;

END END SizeCheckNoDealloc;

(*====================================================================*)
 PROCEDURE SizeCheck(asglen:CARDINAL; VAR s2:DynStr; preserve:BOOLEAN);
(*====================================================================*)

VAR  deallocMe  :DynStr;

BEGIN

SizeCheckNoDealloc(asglen, s2, preserve, deallocMe);

IF deallocMe.str # NIL THEN
   dStrDispose(deallocMe);
END;

END SizeCheck;

(*===========================================*)
 PROCEDURE dStrAsgE(s1:DynStr; VAR s2:DynStr);
(*===========================================*)
   (* Assign one DynStr to an existing one *)

VAR  deallocMe  :DynStr;

BEGIN

SizeCheckNoDealloc(s1.len, s2, s1.str=s2.str, deallocMe);  (* ie, preserve only if in = out *)
DynStr0.dStrAsgE(s1, s2);

IF deallocMe.str # NIL THEN
   dStrDispose(deallocMe);
END;

END dStrAsgE;

(*=================================================*)
 PROCEDURE dStrInE(s1:ARRAY OF CHAR; VAR s2:DynStr);
(*=================================================*)
    (* Assign to an existing DynStr from a "Str0" format string (literal) *)

BEGIN

SizeCheck(LENGTH(s1), s2, FALSE);
DynStr0.dStrInE(s1, s2);

END dStrInE;

(*==================================================*)
 PROCEDURE dStrCat(Suffix:DynStr; VAR OutStr:DynStr);
(*==================================================*)
      (* Concatenate one string onto another *)

VAR  deallocMe  :DynStr;

BEGIN

SizeCheckNoDealloc(OutStr.len+Suffix.len, OutStr, TRUE, deallocMe);
DynStr1.dStrCat(Suffix, OutStr);

IF deallocMe.str # NIL THEN
   dStrDispose(deallocMe);
END;

END dStrCat;

(*==================================================*)
 PROCEDURE dStrPre(Prefix:DynStr; VAR OutStr:DynStr);
(*==================================================*)
      (* Prefix a string onto a string *)

VAR  deallocMe  :DynStr;

BEGIN

SizeCheckNoDealloc(OutStr.len+Prefix.len, OutStr, TRUE, deallocMe);
DynStr1.dStrPre(Prefix, OutStr);

IF deallocMe.str # NIL THEN
   dStrDispose(deallocMe);
END;

END dStrPre;

(*============================================================*)
 PROCEDURE dStr0Pre(Prefix:ARRAY OF CHAR; VAR OutStr:DynStr);
(*============================================================*)

VAR  PrefixLen,
     i,
     CharsToMove    :INTEGER;

BEGIN WITH OutStr DO

PrefixLen := LENGTH(Prefix);

SizeCheck(VAL(INTEGER, len)+PrefixLen, OutStr, TRUE);

CharsToMove := len;

len := PrefixLen + CharsToMove;

i := CharsToMove - 1;
WHILE i >= 0 DO
   str^[i+PrefixLen] := str^[i];
   DEC(i);
END;

CharsToMove := PrefixLen;

i := CharsToMove-1;
WHILE i >= 0 DO
   str^[i] := Prefix[i];
   DEC(i);
END;

str^[len] := 0C;                   (* always available since siz+1 *)

END END dStr0Pre;

(*=================================================*)
 PROCEDURE dStrSeg(InStr:DynStr; Start,Stop:INTEGER;
                   VAR OutStr:DynStr);
(*=================================================*)
      (* Assign a segment of one string to another *)

VAR  needed     :INTEGER;
     deallocMe  :DynStr;

BEGIN

IF Start < 0 THEN
   Start := 0;
END;

IF Stop >= VAL(INTEGER, InStr.len) THEN
   Stop := VAL(INTEGER, InStr.len)-1;
END;

needed := Stop-Start+1;
IF needed < 0 THEN
   needed := 0;
END;

SizeCheckNoDealloc(needed, OutStr, InStr.str = OutStr.str, deallocMe);  (* ie, preserve only if in = out *)
DynStr1.dStrSeg(InStr, Start, Stop, OutStr);

IF deallocMe.str # NIL THEN
   dStrDispose(deallocMe);
END;

END dStrSeg;

(*==========================================================*)
 PROCEDURE dStr0Cat(Suffix:ARRAY OF CHAR; VAR OutStr:DynStr);
(*==========================================================*)
     (* Concatenate one string in "Str0" format onto existing DynStr *)

VAR  i, j       :INTEGER;
     SuffixLen  :CARDINAL;

BEGIN WITH OutStr DO

SuffixLen := LENGTH(Suffix);

SizeCheck(len+SuffixLen, OutStr, TRUE);

j := 0;
FOR i := VAL(INTEGER, len) TO VAL(INTEGER, len+SuffixLen)-1 DO
   str^[i] := Suffix[j];
   INC(j);
END;

len := len + SuffixLen;
str^[len] := 0C;         (* always available since siz+1 *)

END END dStr0Cat;

(*===========================================*)
 PROCEDURE dChrCat(c:CHAR; VAR OutStr:DynStr);
(*===========================================*)
     (* Concatenate one char onto existing DynStr *)

BEGIN

SizeCheck(OutStr.len+1, OutStr, TRUE);
DynStr1C.dChrCat(c, OutStr);

END dChrCat;

(*============================================*)
 PROCEDURE dChrPre(c:CHAR; VAR OutStr:DynStr);
(*============================================*)
     (* Prefix one char onto existing DynStr *)

BEGIN

SizeCheck(OutStr.len+1, OutStr, TRUE);
DynStr1C.dChrPre(c, OutStr);

END dChrPre;

END DynStr3.
