
(*####################################*)
 IMPLEMENTATION MODULE ListV2Processor;    (* v2.2-080102 *)
(*####################################*)
                                       (* by Thomas Breeden, tmb@virginia.edu *)
(*
       This module maintains ordered lists of arbitrary data items.
       Lists are by ADDRESS reference, no copying (or type checking) is done.
*)

(*
Copyright (C) 1985-2002 Thomas Breeden.

Permission to use, copy, modify and distribute this software and its
documentation for any purpose and without fee is hereby granted, provided that
the above copyright notice appear in all copies and that both the copyright
notice and this permission notice appear in supporting documentation.
Thomas Breeden makes no representations about the suitability of this software
for any purpose. It is provided "as is" without express or implied warranty.
*)

FROM SYSTEM     IMPORT ADDRESS, CAST;
FROM Storage    IMPORT ALLOCATE, DEALLOCATE;
FROM Debugging0 IMPORT Assert;

TYPE

   List        = POINTER TO ListContext;

   ListRecPtr  = POINTER TO ListRec;
   ListRec     = RECORD
                    Dat        :ADDRESS;
                    Prv,
                    Nxt        :ListRecPtr;
                  END;

   ListStates  = (Reading, Writing);

   ListContext = RECORD
                     Hdr       :ListRecPtr;
                     CmpProc   :Ordering;
                     DuplOk    :BOOLEAN;
                     curQ,                    (* previous to current pos *)
                     curP      :ListRecPtr;   (* current pos *)
                     State     :ListStates;
                     Done      :BOOLEAN;
                 END;

VAR FatalError      :PROC;
    ListRecPool     :ListRecPtr;
    ListContextPool :List;

(*
NOTE
        1.The lists have an initial (dummy) header record which is pointed
          to by the context block and used simply to make the processing
          code simpler.

        2.Hdr^.Prv points to tail of list. If the list is empty, then
          Hdr^.Prv points to Hdr itself.

        3.When curP is NIL, the position is "beyond the end" and curQ points
          to the last item (or Hdr if an empty list).

          When curP points to Hdr, the position "before the start" and the
          value of curQ is immaterial.

        4.If you allow duplicates, there is NO WAY to insure which one
          of duplicate elements is deleted by ListDeletion or found by
          OnList! However, OnList followed by ListRead(s) is insured to
          process them all.

     In general, if you do not consider "duplicate" elements totally
          interchangeable, you should instead disallow duplicates and write
          your CompareProc to tell them apart.


VERSION HISTORY

v1.0-062685   -First version

v1.1-121085   -ListDestroy now makes the DEALLOCATION of the data elements
               on the list optional.
              -OnList now sets current position to be that found (if found).
              -A new proc, ListRef, similar to ListRead except the current
               position is not stepped afterward.

v1.0-022188 for AmigaDOS
              -New/Dispose -> ALLOCATE/DEALLOCATE.

v2.0-090290   -Added ListReadRev.
              -Added ListEndset.
              -Backpointer added to ListRec. For the list header, this
               points to the last item on the list.
              -Added duplicates allowed list option.
              -BIG CHANGE TO IMPLEMENTATION: Read (and ReadRev) NOW DO THE
               INCREMENT (DECREMENT) CURRENT OPERATION BEFORE READING THE
               DATA RATHER THAN AFTER. THIS MEANS CODE DOING OnList AND THEN
               Read() WILL HAVE TO BE CHANGED. Doing it the old way, however,
               makes changing read directions on the list very awkward.

v2.1-062596   -Now using a recycling Pool of ListRecs and contexts, since the
               ALLOCATE/DEALLOCATE via C lib malloc/free is so slow.

    -042198   -Replaced all the HALTs with Asserts(, "")

v2.2-081298   -Removed ListOpDone as a Global exported variable and replaced it
               with a function call. Now this is maintained on a per list basis.

    -080102   -Implemented NilList() in Stony Brook version.
              -In OnLine() there was a WITH L^ nested immediatedly withing
               another WITH L^. Removed the inner one.
*)

(*------------------------------------*)
 PROCEDURE ObtainListRec():ListRecPtr;
(*------------------------------------*)

VAR  ret :ListRecPtr;

BEGIN

IF ListRecPool = NIL THEN
   ALLOCATE(ret, SIZE(ret^));
ELSE
   ret := ListRecPool;
   ListRecPool := ListRecPool^.Nxt;
END;

RETURN ret;

END ObtainListRec;

(*-------------------------------------*)
 PROCEDURE RecycleListRec(p:ListRecPtr);
(*-------------------------------------*)

BEGIN

p^.Nxt := ListRecPool;
ListRecPool := p;

END RecycleListRec;

(*----------------------------------*)
 PROCEDURE ObtainListContext():List;
(*----------------------------------*)

VAR  ret :List;

BEGIN

IF ListContextPool = NIL THEN
   ALLOCATE(ret, SIZE(ret^));
ELSE
   ret := ListContextPool;
   ListContextPool := CAST(List, ListContextPool^.curP);
END;

RETURN ret;

END ObtainListContext;

(*-----------------------------------*)
 PROCEDURE RecycleListContext(p:List);
(*-----------------------------------*)

BEGIN

p^.curP := CAST(ListRecPtr, ListContextPool);
ListContextPool := p;

END RecycleListContext;

(*====================================*)
 PROCEDURE ListOpDone(L:List):BOOLEAN;
(*====================================*)
BEGIN

RETURN L^.Done;

END ListOpDone;

(*========================*)
 PROCEDURE NilList():List;
(*========================*)

BEGIN

RETURN NIL;

END NilList;

(*=========================================================================*)
 PROCEDURE ListCreate(VAR L:List; CompareProc:Ordering; Duplicates:BOOLEAN);
(*=========================================================================*)

BEGIN

L := ObtainListContext();

WITH L^ DO
   Hdr := ObtainListRec();
   Hdr^.Dat := NIL;
   Hdr^.Nxt := NIL;
   Hdr^.Prv := Hdr;       (* ? *)
   CmpProc := CompareProc;
   DuplOk := Duplicates;
   State := Reading;
   Done := TRUE;
END;
ListReset(L);
ListSetWriteMode(L);

END ListCreate;

(*======================================================================*)
 PROCEDURE ListDestroy(VAR L:List; DataAlso:BOOLEAN; DatumSize:CARDINAL);
(*======================================================================*)

VAR Q :ListRecPtr;

BEGIN WITH L^ DO

State := Reading;
ListReset(L);           (* here curP is Hdr *)
curP := curP^.Nxt;      (* here curP is first elem *)  (* v.2.0 *)

WHILE curP # NIL DO
   IF DataAlso THEN
      DEALLOCATE(curP^.Dat, DatumSize);
   END;
   Q := curP; curP := curP^.Nxt;
   RecycleListRec(Q);
END;

RecycleListRec(Hdr);
RecycleListContext(L);

END END ListDestroy;

(*================================*)
 PROCEDURE ListSetReadMode(L:List);
(*================================*)

BEGIN

L^.State := Reading;
ListReset(L);

END ListSetReadMode;

(*=================================*)
 PROCEDURE ListSetWriteMode(L:List);
(*=================================*)

BEGIN

L^.State := Writing;
L^.Done := TRUE;

END ListSetWriteMode;

(*==========================*)
 PROCEDURE ListReset(L:List);
(*==========================*)

BEGIN WITH L^ DO

Assert(State = Reading, "ListV2Processor: Not Reading in ListReset");      (* before beginning *)

Done := TRUE;
curQ := NIL;
curP := Hdr;

END END ListReset;

(*===========================*)
 PROCEDURE ListEndset(L:List);
(*===========================*)

BEGIN WITH L^ DO

Assert(State = Reading, "ListV2Processor: Not Reading in ListEndset");

Done := TRUE;
curP := NIL;     (* after end of list *)
curQ := Hdr^.Prv;     (* last element *);

END END ListEndset;

(*============================================*)
 PROCEDURE ListRead(L:List; VAR Datum:ADDRESS);
(*============================================*)

BEGIN WITH L^ DO

Assert(State = Reading, "ListV2Processor: Not Reading in ListRead");

IF curP = NIL THEN               (* after list end *)
   Done := FALSE;
   Datum := NIL;
ELSE
   curQ := curP;        (* advance list position *)
   curP := curP^.Nxt;
   IF curP # NIL THEN
      Done := TRUE;
      Datum := curP^.Dat;
   ELSE                          (* reached list end *)
      Done := FALSE;
      Datum := NIL;
   END;
END;

END END ListRead;

(*===============================================*)
 PROCEDURE ListReadRev(L:List; VAR Datum:ADDRESS);
(*===============================================*)

BEGIN WITH L^ DO

Assert(State = Reading, "ListV2Processor: Not Reading in ListReadRev");

IF curP = Hdr THEN               (* before list start *)
   Done := FALSE;
   Datum := NIL;
ELSE
   curP := curQ;        (* retreat list position *)
   curQ := curQ^.Prv;
   IF curP # Hdr THEN
      Done := TRUE;
      Datum := curP^.Dat;
   ELSE                          (* reached list start *)
      Done := FALSE;
      Datum := NIL;
   END;
END;

END END ListReadRev;

(*===========================================*)
 PROCEDURE ListRef(L:List; VAR Datum:ADDRESS);
(*===========================================*)

BEGIN WITH L^ DO

Assert(State = Reading, "ListV2Processor: Not Reading in ListRef");

IF (curP = NIL) OR (curP = Hdr) THEN   (* after list end or before start *)
   Done := FALSE;
   Datum := NIL;
ELSE
   Done := TRUE;
   Datum := curP^.Dat;
END;

END END ListRef;

(*----------------------------------------------------------------------*)
 PROCEDURE SearchList(L:List; Datum:ADDRESS; VAR Q,P:ListRecPtr):BOOLEAN;
(*----------------------------------------------------------------------*)
(*
   RETURNs TRUE/FALSE for match/nomatch and

   Q pointing to previous element, P pointing at match or to
   following element if no match.
*)

VAR Comparison  :INTEGER;

BEGIN WITH L^ DO

Q := Hdr;        (* guarenteed non-NIL *)
P := Q^.Nxt;

LOOP
   IF P = NIL THEN
      RETURN FALSE;
   END;

   Comparison := CmpProc(P^.Dat, Datum);
   IF Comparison > 0 THEN
      RETURN FALSE;
   ELSIF Comparison = 0 THEN
      RETURN TRUE;
   END;
   Q := P;
   P := P^.Nxt;
END;

END END SearchList;

(*=====================================================================*)
 PROCEDURE OnList(L:List; Datum:ADDRESS; VAR ListElem:ADDRESS):BOOLEAN;
(*=====================================================================*)
(*
*)

VAR  FoundRec,
     Qdum      :ListRecPtr;
     Found     :BOOLEAN;

BEGIN WITH L^ DO

Assert(State = Reading, "ListV2Processor: Not Reading in OnList");

Done := TRUE;
Found := SearchList(L, Datum, Qdum, FoundRec);
IF Found THEN
   ListElem := FoundRec^.Dat;
   curQ := Qdum; curP := FoundRec;
ELSE
   ListElem := NIL;
END;

RETURN Found;

END END OnList;

(*============================================*)
 PROCEDURE ListAddition(L:List; Datum:ADDRESS);
(*============================================*)
(*
NOTE    1.Never inserted twice if NOT DuplOk.
        2.Otherwise, duplicates inserted before match!
*)

VAR NewElement,
    P,Q         :ListRecPtr;

BEGIN WITH L^ DO

Assert(State = Writing, "ListV2Processor: Not Writing in ListAddition");

IF SearchList(L, Datum, Q,P) AND (NOT DuplOk) THEN
   Done := FALSE;
   RETURN;
ELSE
   Done := TRUE;
   NewElement := ObtainListRec();
   NewElement^.Dat := Datum;
   NewElement^.Nxt := P;
   NewElement^.Prv := Q;
   Q^.Nxt := NewElement;
   IF P = NIL THEN
      Hdr^.Prv := NewElement;     (* new end item *)
   END;
END;

END END ListAddition;

(*================================================================*)
 PROCEDURE ListUpdate(L:List; Datum:ADDRESS; VAR ListElem:ADDRESS);
(*================================================================*)

VAR  FoundRec,
     Qdum      :ListRecPtr;

BEGIN WITH L^ DO

Assert(State = Writing, "ListV2Processor: Not Writing in ListUpdate");

IF NOT SearchList(L, Datum, Qdum, FoundRec) THEN
   Done := FALSE;
   RETURN;
ELSE
   Done := TRUE;
   ListElem := FoundRec^.Dat;
   FoundRec^.Dat := Datum;
END;

END END ListUpdate;

(*============================================*)
 PROCEDURE ListDeletion(L:List; Datum:ADDRESS);
(*============================================*)

(*
*)

VAR  P,Q  :ListRecPtr;

BEGIN WITH L^ DO

Assert(State = Writing, "ListV2Processor: Not Writing in ListDeletion");

IF NOT SearchList(L, Datum, Q,P) THEN
   Done := FALSE;
   RETURN;
ELSE
   Done := TRUE;
   IF P^.Nxt = NIL THEN     (* new end item *)
      Hdr^.Prv := Q;
   END;
   Q^.Nxt := P^.Nxt;
   RecycleListRec(P);
END;

END END ListDeletion;

(*---------------*)
 PROCEDURE DoHALT;
(*---------------*)

BEGIN

Assert(FALSE, "ListV2Processor: DoHALT");;

END DoHALT;

(*========================================*)
 PROCEDURE ListFatalErrorActionSet(p:PROC);
(*========================================*)

BEGIN

FatalError := p;

END ListFatalErrorActionSet;

(*===================================*)
 PROCEDURE ListFatalErrorActionReset;
(*===================================*)

BEGIN

FatalError := DoHALT;

END ListFatalErrorActionReset;

(*-------------------------*)
 BEGIN (* module init code *)
(*-------------------------*)

ListFatalErrorActionReset;
ListRecPool := NIL;
ListContextPool := NIL;

END ListV2Processor.
