
(*############################*)
 IMPLEMENTATION MODULE TBChans;                       (* v0.2-102503 *)
(*############################*)
                                       (* by Thomas Breeden, tmb@virginia.edu *)
(*
     "Device" for ISO Std Channel I/O to memory.

     This is an initial try at implementing a new ISO channel to use with
     TextIO, WholeIO, etc.

     The implementation is layered on a module, TextBlocks, and the GetTB()
     provides a hook to get the underlying text data.
*)

(*
Copyright (C) 1997-2003 Thomas Breeden.

Permission to use, copy, modify and distribute this software and its
documentation for any purpose and without fee is hereby granted, provided that
the above copyright notice appear in all copies and that both the copyright
notice and this permission notice appear in supporting documentation.
Thomas Breeden makes no representations about the suitability of this software
for any purpose. It is provided "as is" without express or implied warranty.
*)

IMPORT SYSTEM;
FROM Storage IMPORT ALLOCATE, DEALLOCATE;

IMPORT ChanConsts, IOConsts, IOLink, IOChan;

FROM DynStr0    IMPORT DynStr;
FROM TextBlocks IMPORT AllocBlock, AppendBlockChar, DeallocBlock, EmptyBlock,
                       InvalidBlock, NewBlockLine, NextBlockLine, ResetBlockRead,
                       TextBlock;

IMPORT DynStr0, Str0;

CONST CR = 15C;
      LF = 12C;

TYPE CharPointer = POINTER TO CHAR;

TYPE  TBFileRec = RECORD
                    tb             :TextBlock;
                    linebuf        :DynStr;
                    lpos           :INTEGER;   (* -1 means no existing info in linebuf *)
                    NewLinePending :BOOLEAN;
                  END;
      TBFile = POINTER TO TBFileRec;

VAR  TBFiledid  :IOLink.DeviceId;

(*VERSION HISTORY

vx.x-090397  -First

v0.1-031802  -Now supporting Read as well as Write. Had to add a line buffer/pos to record.
             -Initializing the DeviceTable Flags to {read,write,text}
             -Reset() now longer does EmptyBlock(), but ResetBlockRead(). EmptyBlock()
              call moved to Rewrite().
             -Fixed bug: incorrect test for failed TextBlocks.AllocBlock() in OpenTB().
             -Added a GetName proc to return "TBFile".

    -061003  -Changes for Benchmark M2
                 > In GetName() compiler rejected str := "TBFile" as incompatible assignment.
                   Using Str0.StrAsg() instead
                 > Compiler won't unqualifiedly import enum type items when type
                   TYPE OpenResults = ChanConsts.OpenResults in DEF file. Added ChanConsts.
                   explicitly to these references

v0.2-091403  -Fix in Look(). EndOfInput was not being returned correctly.
                   IF f^.lpos >= VAL(INTEGER, f^.linebuf.len) THEN
              was replaced with
                   IF (rRes = IOConsts.allRight) AND (f^.lpos >= VAL(INTEGER, f^.linebuf.len)) THEN
              Also, f^.lpos := 0; is no longer executed if EndOfInput.

    -102503  -Workaround to fix the TextBlocks problem that it always adds an empty line to the
              end. .NewLinePending field used to avoid actually triggering the NewBlockLine()
              until something is output to the last line.
*)

(*-----------------------*)
 PROCEDURE WrongDeviceEx;
(*-----------------------*)
BEGIN

IOLink.RAISEdevException(IOChan.InvalidChan(), TBFiledid, IOChan.notAvailable,
                                        "Not TBFile device");

END WrongDeviceEx;

(*======================================*)
 PROCEDURE IsTBFile(cid:ChanId):BOOLEAN;
(*======================================*)
BEGIN

RETURN IOLink.IsDevice(cid, TBFiledid);

END IsTBFile;

(*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*)
 PROCEDURE LnWrite(p:IOLink.DeviceTablePtr);
(*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*)

VAR   f   :TBFile;

BEGIN

f := p^.cd;

IF f^.NewLinePending THEN
   NewBlockLine(f^.tb);
ELSE
   f^.NewLinePending := TRUE;
END;

END LnWrite;

(*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*)
 PROCEDURE TextWrite(p:IOLink.DeviceTablePtr; txtadr:SYSTEM.ADDRESS; numbytes:CARDINAL);
(*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*)

VAR  cp  :CharPointer;
     ch  :CHAR;
     i   :CARDINAL;
     f   :TBFile;

BEGIN

IF numbytes = 0 THEN
   RETURN;
END;

f := p^.cd;
cp := txtadr;

IF f^.NewLinePending THEN
   NewBlockLine(f^.tb);
   f^.NewLinePending := FALSE;
END;

FOR i := 1 TO numbytes-1 DO
   ch := cp^;
   AppendBlockChar(f^.tb, ch);
   cp := SYSTEM.ADDADR(cp, 1);
END;

ch := cp^;

IF (ch = CR) OR (ch = LF) THEN
   f^.NewLinePending := TRUE;
ELSE
   AppendBlockChar(f^.tb, ch);
END;

END TextWrite;

(*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*)
 PROCEDURE Look(p:IOLink.DeviceTablePtr; VAR ch:CHAR; VAR rRes:IOConsts.ReadResults);
(*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*)

VAR   f   :TBFile;

BEGIN

f := p^.cd;

rRes := IOConsts.allRight;

IF f^.lpos < 0 THEN
   IF NOT NextBlockLine(f^.tb, f^.linebuf) THEN
      rRes := IOConsts.endOfInput;
   ELSE
      f^.lpos := 0;
   END;
END;

IF (rRes = IOConsts.allRight) AND (f^.lpos >= VAL(INTEGER, f^.linebuf.len)) THEN
   rRes := IOConsts.endOfLine;
END;

IF rRes = IOConsts.allRight THEN
   ch := f^.linebuf.str^[f^.lpos];
END;

p^.result := rRes;

END Look;

(*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*)
 PROCEDURE Skip(p:IOLink.DeviceTablePtr);
(*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*)

VAR   f   :TBFile;

BEGIN

f := p^.cd;

IF f^.lpos < 0 THEN
   IF NOT NextBlockLine(f^.tb, f^.linebuf) THEN
      p^.result := IOConsts.endOfInput;
      IOLink.RAISEdevException(p^.cid, p^.did, IOChan.skipAtEnd, "Skip attempted at end of Input");
   END;
   f^.lpos := 0;
END;

IF f^.lpos >= VAL(INTEGER, f^.linebuf.len) THEN
   f^.lpos := -1;                   (* no read-ahead *)
ELSE
   INC(f^.lpos);
END;

p^.result := IOConsts.allRight;

END Skip;

(*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*)
 PROCEDURE SkipLook(p:IOLink.DeviceTablePtr; VAR ch:CHAR; VAR rRes:IOConsts.ReadResults);
(*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*)

BEGIN

Skip(p);
Look(p, ch, rRes);

END SkipLook;

(*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*)
 PROCEDURE TextRead(p:IOLink.DeviceTablePtr; adr:SYSTEM.ADDRESS; maxChars:CARDINAL;
                    VAR charsRead:CARDINAL);
(*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*)

VAR  f    :TBFile;
     ch   :CHAR;
     ptr  :POINTER TO CHAR;
     rRes :IOConsts.ReadResults;

BEGIN

f := p^.cd;

charsRead := 0;
ptr := adr;
Look(p, ch, rRes);
WHILE (maxChars > 0) AND (rRes = IOConsts.allRight) DO

   DEC(maxChars);
   INC(charsRead);

   ptr^ := ch;
   ptr := SYSTEM.ADDADR(ptr, SIZE(CHAR));

   IF maxChars > 0 THEN
      SkipLook(p, ch, rRes);
   ELSE
      Skip(p);
   END;

END;

END TextRead;

(*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*)
 PROCEDURE Reset(p:IOLink.DeviceTablePtr);
(*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*)

VAR   f   :TBFile;

BEGIN

f := p^.cd;
ResetBlockRead(f^.tb);
f^.NewLinePending := FALSE;

END Reset;

(*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*)
 PROCEDURE Free(p:IOLink.DeviceTablePtr);
(*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*)

VAR   f   :TBFile;

BEGIN

f := p^.cd;
DeallocBlock(f^.tb);
DynStr0.dStrDispose(f^.linebuf);
DISPOSE(f);

END Free;

(*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*)
 PROCEDURE GetName(p:IOLink.DeviceTablePtr; VAR str:ARRAY OF CHAR);
(*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*)
BEGIN

Str0.StrAsg("TBFile", str);

END GetName;

(*----------------------------------------------*)
 PROCEDURE AssignProcs(p:IOLink.DeviceTablePtr);
(*----------------------------------------------*)

BEGIN

(* p^.doFlush := Flush; >>> default if do nothing *)

p^.doFree := Free;

p^.doLook := Look;
p^.doSkip := Skip;
p^.doSkipLook := SkipLook;

(*p^.doRawRead := ;*)
p^.doTextRead := TextRead;

(*p^.doRawWrite := ;*)
p^.doTextWrite := TextWrite;
p^.doLnWrite := LnWrite;

p^.doReset := Reset;
(*p^.doFlush := ;*) (* default is no action *)
p^.doGetName := GetName;

END AssignProcs;

(*====================================================*)
 PROCEDURE OpenTB(VAR cid:ChanId; VAR res:OpenResults);
(*====================================================*)

VAR  p   :IOLink.DeviceTablePtr;
     f   :TBFile;
     ok  :BOOLEAN;

BEGIN

IOLink.MakeChan(TBFiledid, cid);
IF cid # IOChan.InvalidChan() THEN

   p := SYSTEM.CAST(IOLink.DeviceTablePtr, cid);

   ok := FALSE;
   NEW(f);
   IF f # NIL THEN
      f^.linebuf.str := NIL;
      f^.lpos := -1;
      f^.NewLinePending := FALSE;
      f^.tb := AllocBlock();
      IF f^.tb # InvalidBlock() THEN
         DynStr0.dStrCreate(f^.linebuf, 80);
         IF f^.linebuf.str # NIL THEN
            ok := TRUE;
         END;
      END;
   END;

   IF NOT ok THEN
      IF f # NIL THEN
         IF f^.linebuf.str # NIL THEN
            DynStr0.dStrDispose(f^.linebuf);
         END;
         DISPOSE(f);
      END;
      IOLink.UnMakeChan(TBFiledid, cid);
      cid := IOChan.InvalidChan();
      res := ChanConsts.otherProblem;
      RETURN;
   END;

   res := ChanConsts.opened;
   p^.cd := f;
   p^.flags := ChanConsts.FlagSet{ChanConsts.readFlag, ChanConsts.writeFlag, ChanConsts.textFlag};
   AssignProcs(p);

ELSE
   res := ChanConsts.outOfChans;
END;

END OpenTB;

(*=============================*)
 PROCEDURE Reread(cid:ChanId);
(*=============================*)

VAR  p : IOLink.DeviceTablePtr;

BEGIN

p := SYSTEM.CAST(IOLink.DeviceTablePtr, cid);

Reset(p);

END Reread;

(*=============================*)
 PROCEDURE Rewrite(cid:ChanId);
(*=============================*)

VAR  f   :TBFile;
     p   :IOLink.DeviceTablePtr;

BEGIN

p := SYSTEM.CAST(IOLink.DeviceTablePtr, cid);

Reset(p);

f := p^.cd;
EmptyBlock(f^.tb);

END Rewrite;

(*==============================*)
 PROCEDURE Close(VAR cid:ChanId);
(*==============================*)

VAR  p : IOLink.DeviceTablePtr;

BEGIN

IF IsTBFile(cid) THEN
   p := SYSTEM.CAST(IOLink.DeviceTablePtr, cid);
   p^.doFree(p);
   IOLink.UnMakeChan(TBFiledid, cid);
   cid := IOChan.InvalidChan();
ELSE
   WrongDeviceEx;
END;

END Close;

(*=====================================*)
 PROCEDURE GetTB(cid:ChanId):TextBlock;
(*=====================================*)

VAR  p   :IOLink.DeviceTablePtr;
     f   :TBFile;

BEGIN

IF NOT IsTBFile(cid) THEN
   WrongDeviceEx;
END;

p := SYSTEM.CAST(IOLink.DeviceTablePtr, cid);
f := p^.cd;
RETURN f^.tb;

END GetTB;

(*---------------*)
 BEGIN (* Module *)
(*---------------*)

IOLink.AllocateDeviceId(TBFiledid);

END TBChans.
