
(* empty line at end of block should not imply a real empty line??? *)

(*################################*)
 IMPLEMENTATION MODULE TextBlocks;        (* $VER: TextBlocks.mod 0.0 (3.4.4) *)
(*################################*)
                                       (* by Thomas Breeden, tmb@virginia.edu *)
(*
     TextBlocks are in-memory, line-oriented, dynamically-extending strings
     of text. They can be written, read, and to some extent edited with this
     module.

     TextBlocks are layered on top of dynamic strings implemented by the
     DynStrX modules.
*)

(*
Copyright (C) 1997-2004 Thomas Breeden.

Permission to use, copy, modify and distribute this software and its
documentation for any purpose and without fee is hereby granted, provided that
the above copyright notice appear in all copies and that both the copyright
notice and this permission notice appear in supporting documentation.
Thomas Breeden makes no representations about the suitability of this software
for any purpose. It is provided "as is" without express or implied warranty.
*)

FROM SYSTEM  IMPORT ADDRESS, ADR, CAST;
FROM Storage IMPORT ALLOCATE, DEALLOCATE;

FROM DynStr0 IMPORT DynStr, DynStrPtr;

IMPORT DynStr0, DynStr1, DynStr3, Str0;
   (* DynStr0/1 dynamically allocate string size, but do not thereafter change
      the size. DynStr3 is needed for dynamically re-allocating the .str size *)

FROM Debugging0 IMPORT Assert;

(* VERSION HISTORY

v0.0 (19.5.97)  - Added ResetBlockRead() call in NewBlockLine() if blk was empty before call.
     (22.5.97)  - Removed ResetBlockRead() call in NewBlockLine() if blk was empty before call.
     (25.6.98)  - Added GoToLine()

      (2.8.98)  - Added a call to GetWriteLine() in InitializeBlock() so that AllocBlock and
                  EmptyBlock start off with an existing empty line. Otherwise it is difficult
                  to create a block with an initial 0 length line.

    (27.10.99)  - GetBlockTxt() now inserts a ' ' between lines if there is none at the end or beginning
                  of the two.

    (15.11.00)  - Added InvalidBlock().

    (18.3.02)   - Changed AllocBlock() not to call InitializeBlock() if the ALLOCATE failed.

    (10.6.03)   - unified StonyBrook and Benchmark versions.

    (16.6.03)   - Added GoToMarkRef()

    (4.7.03)    - Added RefBlockLine() and ReadBlockLine()
                - "Fixed" all the Ref...() procs to return a DynStrPtr, so that they can
                  actually fiddle with the length dynstr in the TextBlock, not just read it.

   (26.7.03)    - Removed requirement, NOT ReadOutOfSync, from RefBlockLine(). ok?
                - NewBlockLine() now sets justread to that line and rdr to beyond end.
                  Now ONLY AppendBlkChar can set ReadOutOfSync.
                - Fix in PrevBlockLineRef()
                - After/BeforeInsertBlkLine() now also reset justread and rdr.

   (11.10.3)    - Added AppendBlockLine()
                - DeleteBlockLine() now resets justread and rdr.
                  justread becomes the next line.

   (13.12.3)    - unified StonyBrook and Benchmark versions.

   (3.4.04)     - Added GetBlockNumLines().
*)

CONST  DEBUG = FALSE;

CONST  CR = 15C;
       LF = 12C;

TYPE   LineP = POINTER TO LineInfo;
       LineInfo  = RECORD
                      lin :DynStr;
                      prv,
                      nxt :LineP;
                   END;

TYPE   TextBlock  = POINTER TO TextBlockInfo;
       TextBlockInfo = RECORD
                         hdr           :LineInfo;
                         rdr           :LineP;
                         justread      :LineP;
                         mrk           :ARRAY MarkType OF LineP;
                         num           :CARDINAL;
                         LastCh        :CHAR;
                         ReadOutOfSync :BOOLEAN;
                       END;

                       (* Rules: 1) hdr is a "dummy" record.
                                 2) hdr.nxt is the head record
                                 3) hdr.prv is the tail record
                                 4) hdr nxt (and prv) point to itself for the empty list.

                                 1) rdr pointing to .hdr (ie AtEndList(rdr)) means "all done"
                                 2) justread can be NIL, meaning there is no just read line.
                                    justread is used for things like InsertBefore(), DeleteLine()
                                 3) read fail due to AtEndList(rdr) sets justread to NIL.
                                 4) DeleteLine() sets justread to NIL

                                 5) ReadOutOfSync is set when a write has caused a state such
                                    that ResetBlockRead() must be called to avoid confusion.

                       *)

VAR    LinePool   :LineInfo;

(*-------------------------*)
 PROCEDURE GetLine():LineP;
(*-------------------------*)

VAR  newline :LineP;

BEGIN

IF LinePool.nxt = NIL THEN
   ALLOCATE(newline, SIZE(LineInfo));
   DynStr0.dStrCreate(newline^.lin, 40);
ELSE
   newline := LinePool.nxt;
   LinePool.nxt := newline^.nxt;
END;

newline^.nxt := NIL;
newline^.prv := NIL;

RETURN newline;

END GetLine;

(*----------------------------------*)
 PROCEDURE RecycleLine(VAR l:LineP);
(*----------------------------------*)

BEGIN

l^.nxt := LinePool.nxt;
LinePool.nxt := l;
DynStr1.dStrClr(l^.lin);

l^.prv := NIL;  (* not really necessary *)
l := NIL;

END RecycleLine;

(*--------------------------------*)
 PROCEDURE InitList(blk:TextBlock);
(*--------------------------------*)

BEGIN WITH blk^ DO

hdr.nxt := ADR(hdr);
hdr.prv := ADR(hdr);

END END InitList;

(*-------------------------------------------*)
 PROCEDURE IsEmptyList(blk:TextBlock):BOOLEAN;
(*-------------------------------------------*)

BEGIN WITH blk^ DO

RETURN CAST(ADDRESS, hdr.nxt) = ADR(hdr);

END END IsEmptyList;

(*---------------------------------------------------*)
 PROCEDURE AtEndList(blk:TextBlock; p:LineP):BOOLEAN;
(*---------------------------------------------------*)

BEGIN

RETURN CAST(ADDRESS, p) = ADR(blk^.hdr);

END AtEndList;

(*---------------------------------------*)
 PROCEDURE HeadList(blk:TextBlock):LineP;
(*---------------------------------------*)

BEGIN

RETURN blk^.hdr.nxt;

END HeadList;

(*---------------------------------------*)
 PROCEDURE TailList(blk:TextBlock):LineP;
(*---------------------------------------*)

BEGIN

RETURN blk^.hdr.prv;

END TailList;

(*----------------------------------*)
 PROCEDURE AddAfterList(p, n:LineP);
(*----------------------------------*)

BEGIN

n^.nxt := p^.nxt;
n^.prv := p;

p^.nxt^.prv := n;
p^.nxt := n;

END AddAfterList;

(*----------------------------------*)
 PROCEDURE AddBeforeList(p, n:LineP);
(*----------------------------------*)

BEGIN

n^.nxt := p;
n^.prv := p^.prv;

p^.prv^.nxt := n;
p^.prv := n;

END AddBeforeList;

(*---------------------------------------------*)
 PROCEDURE AddStackList(blk:TextBlock; n:LineP);
(*---------------------------------------------*)

BEGIN

AddAfterList(ADR(blk^.hdr), n);

END AddStackList;

(*---------------------------------------------*)
 PROCEDURE AddQueueList(blk:TextBlock; n:LineP);
(*---------------------------------------------*)

BEGIN

AddAfterList(TailList(blk), n);

END AddQueueList;

(*------------------------------*)
 PROCEDURE DelElemList(d:LineP);
(*------------------------------*)
         (* NOTE: innocuous for d being hdr of empty list *)
BEGIN

d^.prv^.nxt := d^.nxt;
d^.nxt^.prv := d^.prv;

END DelElemList;

(*------------------------------------------*)
 PROCEDURE GetWriteLine(blk:TextBlock):LineP;
(*------------------------------------------*)

BEGIN

IF IsEmptyList(blk) THEN
   NewBlockLine(blk);
END;

RETURN TailList(blk);

END GetWriteLine;

(*---------------------------------------*)
 PROCEDURE InitializeBlock(blk:TextBlock);
(*---------------------------------------*)
                    (* called by AllocBlock() and EmptyBlock() *)

VAR i :INTEGER;
    l :LineP;

BEGIN WITH blk^ DO

LastCh := 0C;

InitList(blk);

num := 0;

FOR i := 0 TO MAX(MarkType) DO
   mrk[i] := NIL;
END;

l := GetWriteLine(blk);

ResetBlockRead(blk);      (* sets justread to NIL, ReadOutOfSync to FALSE *)

END END InitializeBlock;

(*=================================*)
  PROCEDURE AllocBlock():TextBlock;
(*=================================*)

VAR  blk     :TextBlock;

BEGIN

ALLOCATE(blk, SIZE(TextBlockInfo));
IF blk # NIL THEN
   InitializeBlock(blk);
END;

RETURN blk;

END AllocBlock;

(*==========================================*)
  PROCEDURE DeallocBlock(VAR blk:TextBlock);
(*==========================================*)

BEGIN

EmptyBlock(blk);
DEALLOCATE(blk, SIZE(TextBlockInfo));
blk := NIL;

END DeallocBlock;

(*==================================*)
 PROCEDURE InvalidBlock():TextBlock;
(*==================================*)

BEGIN

RETURN NIL;

END InvalidBlock;

(*====================================*)
  PROCEDURE EmptyBlock(blk:TextBlock);
(*====================================*)

VAR  p, q :LineP;

BEGIN

p := HeadList(blk);

WHILE NOT AtEndList(blk, p) DO
   q := p^.nxt;
   RecycleLine(p);
   p := q;
END;

InitializeBlock(blk);

END EmptyBlock;

(*==================================================*)
 PROCEDURE GetBlockNumLines(blk:TextBlock):CARDINAL;
(*==================================================*)

BEGIN WITH blk^ DO

RETURN num;

END END GetBlockNumLines;

(*====================================*)
 PROCEDURE NewBlockLine(blk:TextBlock);
(*====================================*)
     (* add at the block tail *)

VAR (*WasEmpty,
    WasRdrDone  :BOOLEAN;*)

BEGIN WITH blk^ DO

(*>>>WasEmpty := IsEmptyList(blk);<<<*)
(*WasRdrDone := AtEndList(blk, rdr);*)

AddQueueList(blk, GetLine());

INC(num);
LastCh := 0C;

(*>>>IF WasEmpty THEN<<<*)
(*>>>   ResetBlockRead(blk);<<<*)
(*>>>END;<<<*)

SetEndBlockRead(blk);

(*justread := TailList(blk); <<< done in SetEndBlockRead *)

(*
IF WasRdrDone THEN
   rdr := justread; (*ReadOutOfSync := TRUE;*)
END;
*)

END END NewBlockLine;

(*----------------------------------------------------*)
 PROCEDURE IsEOLchar(blk:TextBlock; ch:CHAR):BOOLEAN;
(*----------------------------------------------------*)

BEGIN

IF ch = CR THEN
   NewBlockLine(blk);
ELSIF ch = LF THEN
   IF blk^.LastCh # CR THEN
      NewBlockLine(blk);
   END;
ELSE
   RETURN FALSE;
END;

RETURN TRUE;

END IsEOLchar;

(*================================================*)
 PROCEDURE AppendBlockChar(blk:TextBlock; ch:CHAR);
(*================================================*)

VAR  l  :LineP;

BEGIN WITH blk^ DO

l := GetWriteLine(blk);

IF NOT IsEOLchar(blk, ch) THEN
   DynStr3.dChrCat(ch, l^.lin);
END;

LastCh := ch;

IF AtEndList(blk, rdr) THEN
   ReadOutOfSync := TRUE;
END;

END END AppendBlockChar;

(*==============================================================*)
 PROCEDURE AppendBlockStr0(blk:TextBlock; VAR str:ARRAY OF CHAR);
(*==============================================================*)

VAR  l    :LineP;
     len  :CARDINAL;
     ch   :CHAR;

BEGIN

len := Str0.StrLen(str);
IF len = 0 THEN
   RETURN;
END;

l := GetWriteLine(blk);

ch := str[len-1];
Str0.StrSet(str, len-1);

DynStr3.dStr0Cat(str, l^.lin);

AppendBlockChar(blk, ch);

END AppendBlockStr0;

(*====================================================*)
 PROCEDURE AppendBlockdStr(blk:TextBlock; txt:DynStr);
(*====================================================*)

VAR  ch    :CHAR;
     l     :LineP;
     len   :CARDINAL;

BEGIN

len := DynStr0.dStrLen(txt);
IF len = 0 THEN
   RETURN;
END;

l := GetWriteLine(blk);

ch := txt.str^[txt.len-1];
DEC(txt.len);

DynStr3.dStrCat(txt, l^.lin);

AppendBlockChar(blk, ch);

END AppendBlockdStr;

(*==========================================================*)
 PROCEDURE AddBlockLine(blk:TextBlock; line:DynStr);
(*==========================================================*)
     (* add at the block tail *)

BEGIN

NewBlockLine(blk);

(* l := TailList(blk);  <<< done in NewBlockLine() *)

DynStr3.dStrAsgE(line, blk^.justread^.lin);

(*blk^.LastCh := ?;*)

END AddBlockLine;

(*========================================*)
  PROCEDURE ResetBlockRead(blk:TextBlock);
(*========================================*)

BEGIN WITH blk^ DO

rdr := HeadList(blk);
justread := NIL;
ReadOutOfSync := FALSE;

END END ResetBlockRead;

(*========================================*)
 PROCEDURE SetEndBlockRead(blk:TextBlock);
(*========================================*)

BEGIN WITH blk^ DO

rdr := ADR(blk^.hdr);
justread := TailList(blk);
ReadOutOfSync := FALSE;

END END SetEndBlockRead;

(*======================================================================================*)
 PROCEDURE GoToLineRef(blk:TextBlock; LineNum:CARDINAL; VAR line:DynStrPtr):BOOLEAN;
(*======================================================================================*)
       (* sequentially, from 0 *)

BEGIN WITH blk^ DO

ResetBlockRead(blk);

WHILE NOT AtEndList(blk, rdr) AND (LineNum > 0) DO
   rdr := rdr^.nxt;
   DEC(LineNum);
END;

IF AtEndList(blk, rdr) THEN
   justread := NIL;
   RETURN FALSE;
END;

line := ADR(rdr^.lin);

justread := rdr;
rdr := rdr^.nxt;

RETURN TRUE;

END END GoToLineRef;

(*===================================================================================*)
 PROCEDURE GoToLine(blk:TextBlock; LineNum:CARDINAL; VAR line:DynStr):BOOLEAN;
(*===================================================================================*)
       (* sequentially, from 0 *)

VAR  refds :DynStrPtr;

BEGIN

IF NOT GoToLineRef(blk, LineNum, refds) THEN
   RETURN FALSE;
END;

DynStr3.dStrAsgE(refds^, line);

RETURN TRUE;

END GoToLine;

(*============================================================*)
 PROCEDURE GetBlockTxt(blk:TextBlock; VAR dtxt:DynStr);
(*============================================================*)

BEGIN WITH blk^ DO

DynStr1.dStrClr(dtxt);
ResetBlockRead(blk);

WHILE NOT AtEndList(blk, rdr) DO

   IF rdr^.lin.len > 0 THEN

      IF (dtxt.len > 0) AND (dtxt.str^[dtxt.len-1] # ' ') AND (rdr^.lin.str^[rdr^.lin.len-1] # ' ') THEN
         DynStr3.dStr0Cat(" ", dtxt);
      END;

      DynStr3.dStrCat(rdr^.lin, dtxt);

   END;

   rdr := rdr^.nxt;

END;
justread := TailList(blk);

END END GetBlockTxt;

(*==============================================================*)
 PROCEDURE RefBlockLine(blk:TextBlock; VAR line:DynStrPtr):BOOLEAN;
(*==============================================================*)

BEGIN WITH blk^ DO

(* ??? Assert(NOT ReadOutOfSync, "RefBlockLine: ReadOutOfSync"); ??? *)

(*
IF AtEndList(blk, rdr) THEN
   justread := NIL;
   RETURN FALSE;
END;
*)

Assert(justread # NIL, "justread NIL in RefBlockLine");

line := ADR(justread^.lin);

RETURN TRUE;

END END RefBlockLine;

(*=================================================================*)
 PROCEDURE NextBlockLineRef(blk:TextBlock; VAR line:DynStrPtr):BOOLEAN;
(*=================================================================*)

BEGIN WITH blk^ DO

Assert(NOT ReadOutOfSync, "NextBlockLineRef: ReadOutOfSync");

IF AtEndList(blk, rdr) THEN
   justread := NIL;
   RETURN FALSE;
END;

line := ADR(rdr^.lin);

justread := rdr;
rdr := rdr^.nxt;

RETURN TRUE;

END END NextBlockLineRef;

(*======================================================================*)
 PROCEDURE NextBlockLine(blk:TextBlock; VAR line:DynStr):BOOLEAN;
(*======================================================================*)

VAR  refds :DynStrPtr;

BEGIN

IF NOT NextBlockLineRef(blk, refds) THEN
   RETURN FALSE;
END;

DynStr3.dStrAsgE(refds^, line);

RETURN TRUE;

END NextBlockLine;

(*=========================================================================*)
 PROCEDURE PrevBlockLineRef(blk:TextBlock; VAR line:DynStrPtr):BOOLEAN;
(*=========================================================================*)

VAR  l     :LineP;

BEGIN WITH blk^ DO

Assert(NOT ReadOutOfSync, "PrevBlockLineRef: ReadOutOfSync");

IF justread = NIL THEN

   IF AtEndList(blk, rdr) THEN              (* after NextBlockLine has failed at end of reading *)
      l := TailList(blk);
   ELSIF rdr # HeadList(blk) THEN
      l := rdr^.prv;                  (* after a DeleteBlockLine() ? *)
   ELSE
      RETURN FALSE;                    (* ? before first read *)
   END;

ELSIF justread = HeadList(blk) THEN  (* no previous *)

   justread := NIL;
   rdr := HeadList(blk);
   RETURN FALSE;

ELSE

   l := justread^.prv;
   rdr := rdr^.prv;       (* assuming rdr leads justread *)

END;

line := ADR(l^.lin);

justread := l;

RETURN TRUE;

END END PrevBlockLineRef;

(*===============================================================*)
 PROCEDURE ReadBlockLine(blk:TextBlock; VAR line:DynStr):BOOLEAN;
(*===============================================================*)

VAR  refds :DynStrPtr;

BEGIN

IF NOT RefBlockLine(blk, refds) THEN
   RETURN FALSE;
END;

DynStr3.dStrAsgE(refds^, line);

RETURN TRUE;

END ReadBlockLine;

(*===============================================================*)
 PROCEDURE PrevBlockLine(blk:TextBlock; VAR line:DynStr):BOOLEAN;
(*===============================================================*)

VAR  refds :DynStrPtr;

BEGIN

IF NOT PrevBlockLineRef(blk, refds) THEN
   RETURN FALSE;
END;

DynStr3.dStrAsgE(refds^, line);

RETURN TRUE;

END PrevBlockLine;

(*================================================*)
 PROCEDURE Mark(blk:TextBlock; WhichMark:MarkType);
(*================================================*)
      (* N marks are available *)

BEGIN WITH blk^ DO

Assert(justread # NIL, "no justread in Mark()");

mrk[WhichMark] := justread;

END END Mark;

(*=========================================================================*)
 PROCEDURE GotoMarkRef(blk:TextBlock; WhichMark:MarkType; VAR line:DynStrPtr);
(*=========================================================================*)

BEGIN WITH blk^ DO

Assert(mrk[WhichMark] # NIL, "Unset mark in GotoMark()");
ResetBlockRead(blk);
rdr := mrk[WhichMark];

Assert(NextBlockLineRef(blk, line), "NextBlockLineRef() FAIL in GotoMarkRef?");

END END GotoMarkRef;

(*============================================================================*)
 PROCEDURE GotoMark(blk:TextBlock; WhichMark:MarkType; VAR line:DynStr);
(*============================================================================*)

VAR  refds :DynStrPtr;

BEGIN

GotoMarkRef(blk, WhichMark, refds);

DynStr3.dStrAsgE(refds^, line);

END GotoMark;

(*===============================================================*)
 PROCEDURE AppendBlockLine(blk:TextBlock; appText:DynStr):BOOLEAN;
(*===============================================================*)
      (* append text to just read line *)

BEGIN WITH blk^ DO

IF justread = NIL THEN
   RETURN FALSE;
END;

IF rdr = justread THEN
   Assert(FALSE, "BUG: rdr = justread in AppendBlockLine()");
END;

DynStr3.dStrCat(appText, justread^.lin);

RETURN TRUE;

END END AppendBlockLine;

(*========================================*)
 PROCEDURE DeleteBlockLine(blk:TextBlock);
(*========================================*)
      (* delete the line just read *)

VAR  i :INTEGER;

BEGIN WITH blk^ DO

IF justread = NIL THEN

   RETURN (*FALSE?*);

END;

IF rdr = justread THEN
   Assert(FALSE, "BUG: rdr = justread in DeleteBlockLine()");
END;

DelElemList(justread);

FOR i := 0 TO MAX(MarkType) DO
   IF mrk[i] = justread THEN
      mrk[i] := NIL;
   END;
END;

RecycleLine(justread);
(* justread := NIL; <<<--- done by RecycleLine() *)

DEC(num);

IF AtEndList(blk, rdr) THEN
   justread := NIL;
ELSE
   justread := rdr;
   rdr := rdr^.nxt;
END;

END END DeleteBlockLine;

(*=============================================================*)
 PROCEDURE BeforeInsertBlockLine(blk:TextBlock; line:DynStr);
(*=============================================================*)
      (* insert a line before the line just read *)

VAR  l   :LineP;

BEGIN WITH blk^ DO

l := GetLine();
DynStr3.dStrAsgE(line, l^.lin);

IF justread # NIL THEN
   AddBeforeList(justread, l);
ELSE
   (* nothing just read, put it at the head? *)
   AddStackList(blk, l);
END;

INC(num);
LastCh := 0C;

justread := l;
rdr := l^.nxt;
ReadOutOfSync := FALSE;

END END BeforeInsertBlockLine;

(*================================================================*)
 PROCEDURE AfterInsertBlockLine(blk:TextBlock; line:DynStr);
(*================================================================*)
      (* insert a line after the line just read *)

VAR  l   :LineP;

BEGIN WITH blk^ DO

l := GetLine();
DynStr3.dStrAsgE(line, l^.lin);

IF justread # NIL THEN
   AddAfterList(justread, l);
ELSE
   (* nothing just read, put it at the head? *)
   AddStackList(blk, l);
END;

INC(num);
LastCh := 0C;

justread := l;
rdr := l^.nxt;
ReadOutOfSync := FALSE;

END END AfterInsertBlockLine;

(*=====================================================*)
 PROCEDURE ReplaceBlockLine(blk:TextBlock; line:DynStr);
(*=====================================================*)
      (* the line just read is replaced *)

BEGIN WITH blk^ DO

IF justread = NIL THEN           (* nothing just read, add one at the tail ? *)

   (*DEBUG*)Assert(FALSE, "NULL justread in ReplaceBlockLine()");
   AddBlockLine(blk, line);
   (*blk^.rdr := blk^.tail;*)   (* ? *)

ELSE

   DynStr3.dStrAsgE(line, justread^.lin);

END;

END END ReplaceBlockLine;

(*----------------------*)
 BEGIN (* mod init code *)
(*----------------------*)

LinePool.nxt := NIL;
LinePool.prv := NIL;

END TextBlocks.

